import { Protocol } from './protocol';

export class Service {
  private protocols: Map<string, Protocol> = new Map();
  private stopped = false;

  constructor(private protoType: any) {}

  // Add a new protocol to this service
  public use(pto: Protocol, autostart = true): Protocol {
    // This service is no longer accepting new protocols
    if (this.stopped) {
      return pto;
    }

    const n = pto.name;
    if (n == null || n === '') {
      throw new Error('Invalid Protocol configuration: Needs an unique name!');
    }

    if (this.protocols.has(n)) {
      return this.protocols.get(n) as Protocol;
    }

    this.protocols.set(n, pto);

    if (autostart) {
      pto.start();
    }

    return pto;
  }

  // Stop a protocol running in this service
  public stop(pto: string | Protocol, force = false): void {
    let proto: Protocol;
    if (typeof pto === 'string') {
      if (!this.protocols.get(pto)) {
        return;
      }
      proto = this.protocols.get(pto) as Protocol;
    } else {
      proto = pto;
      if (!this.protocols.get(pto.name)) {
        // Not part of this service!
        return;
      }
    }
    proto.stop(force);
    this.protocols.delete(proto.name);
  }

  public getProtocolByName(name: string): Protocol | undefined {
    return this.protocols.get(name);
  }

  // Gracefully stops everything running in this service
  public stopAll(): Promise<void> {
    return new Promise((resolve, reject) => {
      if (this.stopped) {
        return resolve();
      }

      let toStop = +this.protocols.size;
      let killed = false;

      // Nothing to stop
      if (toStop === 0) {
        return resolve();
      }

      // If the protocols fail to stop within a time frame, we have to kill
      // them in order to prevent lingering connections.
      const killTimeout = setTimeout(() => {
        killed = true;
        this.die();
        resolve();
      }, 10000);

      // Stop everything and wait for them to announce the fact
      // that they've stopped.
      this.stopped = true;
      for (const [name, proto] of this.protocols) {
        proto.once('stopped', () => {
          if (killed) {
            return;
          }

          toStop--;

          if (toStop <= 0) {
            this.protocols.clear();
            clearTimeout(killTimeout);
            resolve();
          }
        });
        proto.stop();
      }
    });
  }

  // Kills everything running in this service
  public die(): void {
    this.stopped = true;
    for (const [name, proto] of this.protocols) {
      proto.stop(true);
    }
    this.protocols.clear();
  }
}
